<?xml version="1.0"?>
<!DOCTYPE gui SYSTEM "kpartgui.dtd">
<gui name="LokalizeMainWindow" version="11">
    <MenuBar>
        <Menu name="file" index="1">
            <text>&amp;File</text>
            <Action name="file_new" />
            <MergeLocal name="new_merge" />
            <Action name="file_open" />
            <Action name="file_open_recent" />
            <MergeLocal name="open_merge" />
            <Separator />
            <Merge />
            <Separator />
            <Action name="file_quit" />
        </Menu>

        <Merge />
        <Menu name="project" index="4">
            <text>&amp;Project</text>
            <Action name="project_overview" />
            <Action name="project_configure" />
            <Separator />
            <Action name="project_open" />
            <Action name="project_open_recent" />
            <Action name="project_close" />
            <Separator />
            <Action name="project_create" />
            <Action name="project_create_odf" />
            <Merge />
        </Menu>
        <Menu name="tools" index="6">
            <MergeLocal name="tools_spelling_merge" />
            <Separator />
            <Action name="tools_filesearch" />
            <Action name="tools_filesearch_next" />
            <Separator />
            <Action name="tools_widgettextcapture" />
            <Separator />
            <Action name="tools_glossary" />
            <Separator />
            <Action name="tools_tm" />
            <Action name="tools_tm_manage" />
            <Merge />
        </Menu>
        <Menu name="settings" index="7">
            <Action name="settings_style" />
            <Action name="colorscheme_menu" />
            <Menu name="toolviews">
                <text>Tool&amp;views</text>
            </Menu>
            <Merge />
        </Menu>
        <Menu name="edit" index="2">
            <text>&amp;Edit</text>
            <MergeLocal />
        </Menu>
        <Menu name="go" index="3">
            <text>&amp;Go</text>
            <MergeLocal />
        </Menu>
        <Menu name="merge" index="5">
            <text>S&amp;ync</text>
            <MergeLocal />
        </Menu>
    </MenuBar>
</gui>
